local LrApplication = import "LrApplication"
local LrFunctionContext = import "LrFunctionContext"
local LrTasks = import "LrTasks"
local LrLogger = import 'LrLogger'
local LrSocket = import "LrSocket"
local LrDialogs = import 'LrDialogs'
local LrDevelopController = import 'LrDevelopController'
local LrApplicationView = import 'LrApplicationView'

local MAIN_LOOP_INTERVAL = 1

local developStates = {}
local developOptions = {
  "Temperature",
  "Tint",
  "Exposure",
  "Highlights", 	
  "Shadows", 
  "Brightness", 
  "Contrast",
  "Whites", 	
  "Blacks",
  "Texture",
  "Clarity",
  "Dehaze",
  "Vibrance",
  "Saturation",
}

local logger = import 'LrLogger'
myLogger = logger( 'LrTpLog' ) -- the log file name.
myLogger:enable( "logfile" )

local function getPhotoJson()
  local catalog = LrApplication.activeCatalog()
  local photo = catalog:getTargetPhoto()

  json = "{"
  if photo:getRawMetadata("path") ~= nil then
    json = json .. "\"path\":\"" .. photo:getRawMetadata("path"):gsub("\\", "\\\\") .. "\","
  end
  if photo:getRawMetadata("rating") ~= nil then
    json = json .. "\"rating\":" .. photo:getRawMetadata("rating") .. ","
  end
  if photo:getRawMetadata("width") ~= nil then
    json = json .. "\"width\":" .. photo:getRawMetadata("width") .. ","
  end
  if photo:getRawMetadata("height") ~= nil then
    json = json .. "\"height\":" .. photo:getRawMetadata("height") .. ","
  end
  if photo:getRawMetadata("aspectRatio") ~= nil then
    json = json .. "\"aspectRatio\":" .. photo:getRawMetadata("aspectRatio") .. ","
  end
  if photo:getRawMetadata("shutterSpeed") ~= nil then
    json = json .. "\"shutterSpeed\":" .. photo:getRawMetadata("shutterSpeed") .. ","
  end
  if photo:getRawMetadata("isoSpeedRating") ~= nil then
    json = json .. "\"isoSpeedRating\":" .. photo:getRawMetadata("isoSpeedRating") .. ","
  end
  if photo:getRawMetadata("aperture") ~= nil then
    json = json .. "\"aperture\":" .. photo:getRawMetadata("aperture") .. ","
  end
  if photo:getRawMetadata("focalLength") ~= nil then
    json = json .. "\"focalLength\":" .. photo:getRawMetadata("focalLength") .. ","
  end
  if photo:getRawMetadata("dateTimeOriginal") ~= nil then
    json = json .. "\"dateTimeOriginal\":\"" .. photo:getRawMetadata("dateTimeOriginal") .. "\","
  end
  if photo:getRawMetadata("isVirtualCopy") ~= nil then
    if photo:getRawMetadata("isVirtualCopy") == true then
      json = json .. "\"isVirtualCopy\":true,"
    else
      json = json .. "\"isVirtualCopy\":false,"
    end
  end
  if photo:getRawMetadata("fileFormat") ~= nil then
    json = json .. "\"fileFormat\":\"" .. photo:getRawMetadata("fileFormat") .. "\","
  end
  
  if photo:getRawMetadata("gps") ~= nil then
    gpsTable = photo:getRawMetadata("gps")
    for i, v in next, gpsTable do
      json = json .. "\""..i.."\":\"" .. v .. "\","
    end
  end
  if photo:getRawMetadata("gpsAltitude") ~= nil then
    json = json .. "\"gpsAltitude\":\"" .. photo:getRawMetadata("gpsAltitude") .. "\","
  end

  if photo:getRawMetadata("keywords") ~= nil then
    keywords = photo:getRawMetadata("keywords")
    keywordsString = ""
    for i, v in next, keywords do
      keywordsString = keywordsString .. v:getName() .. ","
      for n, k in next, v do
        keywordsString = keywordsString .. n .. ","
      end
    end
    json = json .. "\"keywords\":\"" .. keywordsString .. "\","
  end

  json = json .. "\"dummy\":\"1\""
  json = json .. "}"

  -- myLogger:info("[INFO]: Selected Photo: " .. json)
  return json
end

local function processImageAdjustments(message) 

  base = message:gsub("IA", "")
  sep = base:find("|")
  iaType = base:sub(1, sep-1)
  iaVal = base:gsub(iaType.."|", "")

  --LrDialogs.message("Processing", "Applying "..iaType.." with value "..iaVal)
  LrDevelopController.setValue(iaType, tonumber(iaVal) )

end

local function processResetAdjustment(message)
  command = message:gsub("RA|", "")
  myLogger:info("[ACTION]: Resetting adjustment: " .. command)
  LrDevelopController.resetToDefault(command)
end

local function processSwitchModule(message)
  command = message:gsub("SM|", "")
  myLogger:info("[ACTION]: Switch to Module: " .. command)
  LrApplicationView.switchToModule(command)
end

local function processSetDevelopTool(message)
  command = message:gsub("SDT|", "")
  myLogger:info("[ACTION]: Set Develop Tool: " .. command)
  LrDevelopController.selectTool(command)
end

local function processAddKeyword(message)
  local catalog = LrApplication.activeCatalog()
  local command = message:gsub("PAK|", "")
  local result = catalog:withWriteAccessDo( "test" , function() 
    myLogger:info("[ACTION]: Adding Keyword: " .. command)
    local keyword = catalog:createKeyword(command, {}, true, nil, true)
    catalog:getTargetPhoto():addKeyword(keyword)
    end, 
      {
        timeout = 5, 
        callback = function() 
          myLogger:info("[ACTION]: Add Keyword Failed...")
          end
      }
    )
end

local function processRemoveKeyword(message)
  local catalog = LrApplication.activeCatalog()
  command = message:gsub("PRK|", "")
  myLogger:info("[ACTION]: Remove Keyword: " .. command)
  local catalog = LrApplication.activeCatalog()
  local photo = catalog:getTargetPhoto()
  catalog:withWriteAccessDo("actionRemoveKeyword", function(  ) 
    local keyword = catalog:createKeyword(command, {}, true, nil, true)
    photo:removeKeyword(keyword)
  end, {timeout = 5} ) 
end

local function processSetFullscreenMode(message)
  command = message:gsub("SFM|", "") --prev, hide
  if command:find("prev") ~= nil then
    myLogger:info("[ACTION]: Set Fullscreen Preview")
    LrApplicationView.fullscreenPreview()
  elseif message:find("hide") ~= nil then
    myLogger:info("[ACTION]: Set Fullscreen Hide Panels")
    LrApplicationView.fullscreenHidePanels()
  end  
end

local function processZoomAction(message)
  command = message:gsub("ZIO|", "") --zoominnormal, zoominsome, zoomoutnormal, zoomoutsome, 100
  if command:find("zoominnormal") ~= nil then
    myLogger:info("[ACTION]: Zoom In")
    LrApplicationView.zoomIn()
  elseif message:find("zoominsome") ~= nil then
    myLogger:info("[ACTION]: Zoom In Some")
    LrApplicationView.zoomInSome()  
  elseif message:find("zoomoutnormal") ~= nil then
    myLogger:info("[ACTION]: Zoom Out")
    LrApplicationView.zoomOut() 
  elseif message:find("zoomoutsome") ~= nil then
    myLogger:info("[ACTION]: Zoom Out Some")
    LrApplicationView.zoomOutSome()    
  elseif message:find("100") ~= nil then
    myLogger:info("[ACTION]: Zoom to One to One")
    LrApplicationView.zoomToOneToOne()
  end
end

--------------------------------------------------------------------------------
local function connectIncoming( context )
  
  if _G.socketIn~=nill then
    myLogger:info("[Connection State IN]: Reconnecting...")
    _G.socketIn:reconnect()
  else
    myLogger:info("[Connection State IN ]: Binding Socket...")
    _G.socketIn = LrSocket.bind {  functionContext = context,
    port = 14157,
    mode = "receive",
    plugin = _PLUGIN,
    onConnecting = function( socket, port )
      myLogger:info("[Connection State IN ]: Connecting...")
      _G.connectionInState = 1
    end,
    onConnected = function( socket, port )
      myLogger:info("[Connection State IN ]: Connected!")
      _G.connectionInState = 2
    end,
    onClosed = function( socket )
      myLogger:info("[Connection State IN ]: Closed!")
      _G.connectionInState = 1
      socket:reconnect()
    end,
    onMessage = function( socket, message )
      myLogger:info("Incoming message: " .. message)
      if message:find("TPPING2") ~= nil then
        _G.socketOut:send("TPPONG2\n")
        _G.lastPing = os.clock()
        myLogger:info("SEnd pong 2: ")
      elseif message:find("TPPING") ~= nil then
        _G.socketOut:send("TPPONG\n")
      elseif message:find("^IA") ~= nil then
        processImageAdjustments(message)
      elseif message:find("^SM") ~= nil then
        processSwitchModule(message)
      elseif message:find("^SFM") ~= nil then
        processSetFullscreenMode(message)
      elseif message:find("^TZ") ~= nil then
        myLogger:info("[ACTION]: Toggle Zoom")
        LrApplicationView.toggleZoom() 
      elseif message:find("^ZIO") ~= nil then
        processZoomAction(message)
      elseif message:find("^SDT") ~= nil then
        processSetDevelopTool(message)
      elseif message:find("^SDAT") ~= nil then
        myLogger:info("[ACTION]: Apply Develop Auto Tone")
        LrDevelopController.setAutoTone()
      elseif message:find("^SDAWB") ~= nil then
        myLogger:info("[ACTION]: Apply Develop Auto White Balance")
        LrDevelopController.setAutoWhiteBalance()
      elseif message:find("^RAT") ~= nil then
        myLogger:info("[ACTION]: Reset Transforms")
        LrDevelopController.resetTransforms()
      elseif message:find("^RAA") ~= nil then
        myLogger:info("[ACTION]: Reset All Adjustments")
        LrDevelopController.resetAllDevelopAdjustments()
      elseif message:find("^RA") ~= nil then
        processResetAdjustment(message)
      elseif message:find("^PCS") ~= nil then
        -- Copy photo settings
        myLogger:info("[ACTION]: Copy Photo Settings")
        local catalog = LrApplication.activeCatalog()
        local photo = catalog:getTargetPhoto()
        photo:copySettings()
      elseif message:find("^PPS") ~= nil then
        -- Paste photo settings
        myLogger:info("[ACTION]: Paste Photo Settings")
        local catalog = LrApplication.activeCatalog()
        local photo = catalog:getTargetPhoto()
        photo:pasteSettings()
      elseif message:find("^PAK") ~= nil then
        LrTasks.startAsyncTask( function() 
          processAddKeyword(message)
        end)
      elseif message:find("^PRK") ~= nil then
        LrTasks.startAsyncTask( function() 
          processRemoveKeyword(message)
        end)

      end

    end,
    onError = function( socket, err )
      if err == "timeout" then
        myLogger:info("[Connection State IN ]: Timout, reconnecting...")
      else
        myLogger:info("[Connection State IN ]: Error: " .. err .. ", reconnecting...")
      end
      _G.connectionInState = 1
      _G.lastPing = os.clock()
      socket:reconnect()
    end,
  }
  end
  myLogger:info("[Connection State IN]: END connect FUNCTION...")
end

function closeAndReconnectOut()
  _G.socketOut:close()
  _G.socketOut:reconnect()
end

function recreateOut( context )
  myLogger:info("[Connection State OUT]: Binding socket...")
  _G.socketOut = LrSocket.bind {  functionContext = context,
    port = 14158,
    mode = "send",
    plugin = _PLUGIN,
    onConnecting = function( socket, port )
      myLogger:info("[Connection State OUT]: Connecting...")
      -- _G.lastPing = os.clock()
      _G.connectionOutState = 1
    end,
    onConnected = function( socket, port )
      myLogger:info("[Connection State OUT]: Connected!")
      -- _G.lastPing = os.clock()
      _G.connectionOutState = 2
    end,
    onClosed = function( socket )
      myLogger:info("[Connection State OUT]: Closed!")
      _G.connectionOutState = 1
      -- _G.socketOut:reconnect()
      _G.socketOut = nil
      LrTasks.pcall(recreateOut)
    end,
    onMessage = function( socket, message )
      -- Should not happen
    end,
    onError = function( socket, err )
      if err == "timeout" then
        myLogger:info("[Connection State OUT]: Timout, reconnecting...")
      else
        myLogger:info("[Connection State OUT]: Error: " .. err .. ", reconnecting...")
      end
      _G.connectionOutState = 1
      _G.lastPing = os.clock()
      _G.socketOut:reconnect()
      
    end,
  }
end

local function connectOutgoing( context )
  -- myLogger:info("[Connection State OUT]: Init connect...")
  if _G.socketOut~=nill then
    myLogger:info("[Connection State OUT]: Reconnecting...")
    -- _G.socketOut:reconnect()
    LrTasks.pcall(closeAndReconnectOut)
  else
    recreateOut(context)
  end
  myLogger:info("[Connection State OUT]: END connect FUNCTION...")
end

local function updateDevelopChanges()

  for key, value in pairs(developOptions) do
    oldValue = developStates[value]
    developStates[value] = LrDevelopController.getValue(value)
    if oldValue ~= developStates[value] then
      _G.socketOut:send("IAc|"..value.."|"..developStates[value].."\n")
    end
  end

end

local function setupLrCallbacks(context) 

  for key, value in pairs(developOptions) do
    developStates[value] = LrDevelopController.getValue(value)
  end

  LrDevelopController.addAdjustmentChangeObserver(context, "", function(observer)
      updateDevelopChanges()
    end )

end

-- Main Entry Point, our loop
LrTasks.startAsyncTask( function()
    LrFunctionContext.callWithContext( 'mainLoop', function( context )
        myLogger:info("[Initializing]: Starting main loop")
        _G.lastPing = os.clock()
        
        _G.currentModule = "none"
        _G.currentCatalog = "none"
        _G.currentPhotoInfo = "none"

        setupLrCallbacks(context)
        myLogger:info("[Initializing]: Callbacks created")
        _G.connectionInState = 0
        _G.connectionOutState = 0
        _G.isRunning = true
        while _G.isRunning do
          -- myLogger:info("[Main Loop]: Start")
          if _G.connectionOutState == 0 then
            _G.connectionOutState = 1
            myLogger:info("[Main Loop]: Start connection Outgoing...")
            connectOutgoing(context)
          end
          if _G.connectionInState == 0 then
            _G.connectionInState = 1
            myLogger:info("[Main Loop]: Start connection Incoming...")
            connectIncoming(context)
          end

          tmp = _G.lastPing + 10
          if tmp < os.clock() then
            myLogger:info("[Main Loop]: Connection Timeout detected, resetting connection...")
            myLogger:info("[Main Loop]: Current Clock = "..os.clock())
            myLogger:info("[Main Loop]: Temp = "..tmp)
            -- So we are timed out currently so we need to reset connections and such
            
            _G.connectionOutState = 1
            connectOutgoing(context)

            _G.connectionInState = 1
            connectIncoming(context)
            
            _G.lastPing = os.clock()
          end

          -- myLogger:info("[Main Loop]: Checking state changes")
          if _G.connectionOutState == 2 then
            
            -- Module change event
            module = LrApplicationView.getCurrentModuleName()
            if _G.currentModule ~= module then
              _G.socketOut:send("MOD|"..module.."\n")
              _G.currentModule = module
            end

            -- Active catalog event
            local catalog = LrApplication.activeCatalog()
            if _G.currentCatalog ~= catalog:getPath() then
              _G.socketOut:send("E_CAT|"..catalog:getPath().."\n")
              _G.currentCatalog = catalog:getPath()
            end

            local photoInfo = getPhotoJson()
            if _G.currentPhotoInfo ~= photoInfo then
              _G.socketOut:send("E_PI|"..photoInfo.."\n")
              _G.currentPhotoInfo = photoInfo
              LrFunctionContext.callWithContext( 'InitSetKeyWord', function( context )
                processAddKeyword("Dit is een test")
              end)
            end
          end

          -- myLogger:info("[Main Loop]: End")
          LrTasks.sleep( MAIN_LOOP_INTERVAL )
        end

        myLogger:info("[Main Loop]: Stopped")
        if _G.connectionInState > 0 then
          _G.socketIn:close()
          _G.connectionInState = 0
        end
        if _G.connectionOutState > 0 then
          _G.socketOut:close()
          _G.connectionOutState = 0
        end
      end )
  end )